import glob
import os
import re
import shutil
import sys
from concurrent import futures
from subprocess import PIPE, Popen  # execution
from pathlib import Path

import sltxpkg.globals as sg
from sltxpkg.config import load_dependencies_config, write_to_log
from sltxpkg.globals import (C_AUTODETECT_DRIVERS, C_CLEANUP, C_CREATE_DIRS,
							 C_DRIVER_LOG, C_DRIVER_PATTERNS, C_DRIVERS,
							 C_RECURSIVE, C_TEX_HOME, C_DOWNLOAD_DIR,
							 DEFAULT_CONFIG, print_idx)
import sltxpkg.util as su
from sltxpkg import dep

loaded = []


def detect_driver(idx: str, url: str):
	print_idx(idx, " - Autodetecting driver...")
	for key, patterns in sg.configuration[C_DRIVER_PATTERNS].items():
		for pattern in patterns:
			if re.search(pattern, url):
				return key
	print_idx(idx, " ! No driver found...")
	sys.exit(1)


def split_grab_pattern(pattern: str, default_target: str) -> (str, str):
	parts = pattern.split('=>', 1)
	return (parts[0], default_target if len(parts) == 1 else parts[1])


def grab_from(idx: str, path: str, data: dict, target: str, key: str, grabber) -> bool:
	if key not in data:
		print_idx(idx, " ! Key '" + key + "' not found. Won't grab any...")
		return False

	grabs = []
	for grab_pattern in data[key]:
		pattern = split_grab_pattern(grab_pattern, target)
		# maybe forbid level up?
		grabs.extend(map(lambda x, pattern=pattern: (x, pattern[1]), glob.glob(os.path.join(
			path, pattern[0]), recursive=True)))

	# TODO: rel path for files?
	# extra so i can setup installer afterwards more easily
	print_idx(idx, " > Grabbing the follwing for installation:",
			  [os.path.relpath(f[0], path) for f in grabs])
	for grab in grabs:
		grabber(grab, target, path)
	return True


def f_grab_files(data: (str, str), target: str, path: str):
	file_target = os.path.join(target, data[1]) if data[1] != target else os.path.join(
		data[1], os.path.relpath(data[0], path))
	Path(file_target).parent.mkdir(parents=True, exist_ok=True)
	shutil.copy2(data[0], file_target)


def f_grab_dirs(data: (str, str), target: str, path: str):
	if sys.version_info < (3, 8, 0):
		print("Python version below 3.8, falling back with distutils!")
		import distutils.dir_util as du

	# only choose relative path
	dir_target = os.path.join(target, data[1]) if data[1] != target else os.path.join(
		data[1], os.path.relpath(data[0], path))
	Path(dir_target).parent.mkdir(parents=True, exist_ok=True)
	if sys.version_info >= (3, 8, 0):  # we have exist is ok
		shutil.copytree(data[0], dir_target, dirs_exist_ok=True)
	else:
		du.copy_tree(data[0], dir_target)


def write_proc_to_log(idx: int, stream, mirror: bool):
	while True:
		line = stream.readline()
		if not line:
			break
		line_utf8 = line.decode('utf-8')
		write_to_log(line_utf8)
		if mirror:
			print_idx(idx, line_utf8)


def grab_stuff(idx: str, dep_name: str, target_dir: str, data: dict, target: str):
	print_idx(idx, " > Grabbing dependencies for " + dep_name)
	print_idx(idx, "   - Grabby-Grab-Grab files from \"" + target_dir + "\"...")
	got_files = grab_from(idx, target_dir, data, target,
						  'grab-files', f_grab_files)
	print_idx(idx, " - Grabby-Grab-Grab dirs from \"" + target_dir + "\"...")
	got_dirs = grab_from(idx, target_dir, data, target,
						 'grab-dirs', f_grab_dirs)
	if not got_files and not got_dirs:
		print_idx(idx, " ! No grabs performed!")
		write_to_log("No grabs performed for: " + dep_name)


def get_target_dir(data: dict, dep_name: str, driver: str):
	return sg.configuration[C_DRIVERS][driver]["target-dir"].format(
		**data, **sg.configuration, dep_name=dep_name)


def recursive_dependencies(idx: str, driver_target_dir: str, data: dict, dep_name: str, target: str):
	if 'dep' not in data:
		print_idx(idx, "No 'dep' key found for dep: " + dep_name +
				  " using the default (" + sg.DEFAULT_DEPENDENCY + ")")
		data['dep'] = sg.DEFAULT_DEPENDENCY
	dep_files = glob.glob(os.path.join(
		driver_target_dir, data['dep']), recursive=True)
	print_idx(idx, " - Found dep-config:", dep_files)

	if len(dep_files) <= 0:
		return

	new_dependencies = {}
	for dep_file in dep_files:
		new_dependencies = load_dependencies_config(dep_file, new_dependencies)

	_install_dependencies(idx, new_dependencies, target)


def use_driver(idx: str, data: dict, dep_name: str, driver: str, url: str, target: str):
	# default no arguments
	if "args" not in data:
		data["args"] = ""
	driver_data = sg.configuration[C_DRIVERS][driver]
	command = driver_data["command"].format(
		**data, **sg.configuration, dep_name=dep_name)
	driver_target_dir = get_target_dir(data, dep_name, driver)
	if os.path.isdir(driver_target_dir) and driver_data["needs-delete"]:
		print_idx(idx, " - Target folder", driver_target_dir,
				  "exists. Will be deleted as the driver needs this")
		shutil.rmtree(driver_target_dir)
	print_idx(idx, " > Executing:", command)
	feedback = Popen(command, stdout=PIPE, stderr=PIPE, shell=True)
	return_code = feedback.wait()
	write_proc_to_log(idx, feedback.stdout, False)
	if return_code != 0:
		print_idx(idx, " - Error-Log of Driver:")
	write_proc_to_log(idx, feedback.stderr, return_code != 0)

	if(sg.configuration[C_RECURSIVE]):
		recursive_dependencies(idx, driver_target_dir, data, dep_name, target)

	if return_code != 0:
		print_idx(idx, " ! Driver failed with code", feedback, "exiting.")
		sys.exit(return_code)

	grab_stuff(idx, dep_name, driver_target_dir, data, target)


def install_dependency(name: str, idx: str, data: dict, target: str):
	print_idx(idx, "Loading \"" + name + "\"")
	if "url" not in data:
		print_idx(idx, " ! The dependency did not have an url-tag attached")
	url = data["url"]
	print_idx(idx, " - Loading from: \"" + url + "\"")
	if "driver" not in data:
		if not sg.configuration[C_AUTODETECT_DRIVERS]:
			print_idx(idx, " ! No driver given and autodetection disabled!")
		else:
			data["driver"] = detect_driver(idx, url)
	driver = data["driver"]
	print_idx(idx, " - Using driver: \"" + driver + "\"")

	if name in loaded:
		print_idx(idx, " > Skipping retrieval", name,
				  " as it was already loaded by another dep.")
		grab_stuff(idx, name, get_target_dir(data, name, driver), data, target)
		return
	loaded.append(name)

	if driver not in sg.configuration[C_DRIVERS]:
		print_idx(idx, " ! The selected driver is unknown. Loaded:",
				  sg.configuration[C_DRIVERS])
		sys.exit(2)

	use_driver(idx, data, name, driver, url, target)


def _install_dependencies(idx: int, dep_dict: dict, target: str, first: bool = False):
	with futures.ThreadPoolExecutor(max_workers=sg.args.threads) as pool:
		runners = []
		for i, dep in enumerate(dep_dict['dependencies']):
			runners.append(pool.submit(install_dependency, dep, str(
				i) if first else str(idx) + "." + str(i), dep_dict['dependencies'][dep], target))
		futures.wait(runners)
		for runner in runners:
			if runner.result() is not None:
				print(runner.result())


def install_dependencies(target: str = su.get_sltx_tex_home()):
	if "target" not in sg.dependencies or "dependencies" not in sg.dependencies:
		print("The dependency-file must supply a 'target' and an 'dependencies' key!")
		sys.exit(1)

	write_to_log("====Dependencies for:" + sg.dependencies["target"]+"\n")
	print()
	print("Dependencies for:", sg.dependencies["target"])
	print("Installing to:", target)
	print()

	_install_dependencies(0, sg.dependencies, target, first=True)

	# all installed
	if sg.configuration[C_CLEANUP]:
		print("> Cleaning up the download directory, as set.")
		shutil.rmtree(sg.configuration[C_DOWNLOAD_DIR])
	print("Loaded:", dep.loaded)
	if not sg.configuration[C_RECURSIVE]:
		print("Recursion was disabled.")
	print("Dependency installation for",
		  sg.dependencies["target"], "completed.")




def add_argument_to_parser(p, *args, **kwargs):
	p.add_argument(*args, **kwargs)


def add_parser(p, name, arguments, **kwargs):
	np = p.add_parser(name, **kwargs)
	for args in arguments:
		add_argument_to_parser(np, *args.args, **args.kwargs)


class Arg:
	def __init__(self, *args, **kwargs):
		self.args = args
		self.kwargs = kwargs


class Commands:
	def __init__(self, cmds:dict):
		self.cmds = cmds

	def helper_values(self):
		return map(lambda x: x[0], self.cmds.values())

	def generate(self, p):
		for command in self.cmds.items():
			add_parser(p, command[0], command[1][2], **((dict(command[1][1].kwargs, aliases=command[1][0][1])) if len(command[1][0][1]) != 0 else command[1][1].kwargs))


# Docker control
import json
import os
import sys

import docker
import sltxpkg.globals as sg
from sltxpkg.globals import DOCKER_URL
import sltxpkg.util as su

class DockerCtrl:
	def __init__(self):
		self.client = docker.from_env()
		self.client.ping()

	def update_img(self, profile: str):
		target = DOCKER_URL.format(**locals())
		print("Pulling image:", target, "this may take a few minutes")
		for line in self.client.api.pull(target, tag='latest', stream=True):
			# default values
			line = line.decode('utf-8')
			lines = line.split('\r\n')
			for subline in lines:
				if subline is None or subline.strip() == "":
					continue
				d = {'status': 'unknown', 'progress': '', 'id': ''}
				d = {**d, **json.loads(subline)}
				print("   {status} {progress} {id}".format(**d))
	# tODO: autocache name with unique path needs to use real file name not docker!!
	def run_in_container(self, root: bool, profile: str, command: str):
		if profile.startswith(":"):
			target = profile[1:]
		else:
			target = DOCKER_URL.format(**locals())
		print("Launching container based on image:", target)
		if root:
			print("Using root configuration. This might lead to permission errors in the future.", target)
		# TODO: this must be expanded better and safer, this way only '~' might be used which is bad
		wd = sg.configuration[sg.C_WORKING_DIR].replace(os.path.expanduser('~'), '/root')
		print("  - Note: Working-Dir bound to:", wd,"for",sg.configuration[sg.C_WORKING_DIR])
		print("  - Note: Main-Dir bound to: /root/data for",os.getcwd())
		volumes = {
				os.getcwd(): {
					'bind': '/root/data',
					'mount': 'rw'
				},
				sg.configuration[sg.C_WORKING_DIR]: {
					'bind': wd,
					'mount': 'rw'
				}
			}
		if sg.args.local_texmf:
			target_mount = "/usr/share/sltx/texmf"
			print("  - Note: Mounting local txmf-tree (" + su.get_tex_home() + ") to",target_mount)
			volumes[su.get_tex_home()] = {
				'bind': target_mount,
				'mount': 'rw'
			}
		run = self.client.containers.run(
			target, command=command, detach=True, remove=True, working_dir='/root/data',tty=True,
			network_mode='bridge',user='root' if root else 'lithie-user',
			volumes=volumes)
		# We need a buffer in case a multibyte unicode sequence
		# will be broken at line end
		buffer = b''
		for l in run.logs(stdout=True, stderr=True, stream=True, timestamps=True):
			try:
				print((buffer + l).decode('utf-8'), end='')
				buffer = b''
			except UnicodeDecodeError as ex:
				buffer += l
		print("Container completed.")
		feedback = run.wait()
		if 'StatusCode' in feedback and feedback['StatusCode'] != 0:
			code = feedback['StatusCode']
			print("Command failed with:",code)
			sys.exit(code)


import os
import sys

import yaml

from sltxpkg import globals as sg
from sltxpkg import util as su
from sltxpkg.globals import (C_AUTODETECT_DRIVERS, C_CLEANUP, C_CREATE_DIRS,
							 C_DOWNLOAD_DIR, C_DRIVER_LOG, C_DRIVER_PATTERNS,
							 C_DRIVERS, C_TEX_HOME, C_WORKING_DIR, C_CACHE_DIR,
							 DEFAULT_CONFIG)


def write_to_log(data: str):
	if sg.configuration[C_DRIVER_LOG].strip():
		with open(sg.configuration[C_DRIVER_LOG], 'a') as f:
			f.write(data)
			if not data.endswith('\n'):
				f.write("\n")


def load_configuration(file: str):
	y_conf = su.load_yaml(file)
	sg.configuration = {**sg.configuration, **y_conf}


def load_dependencies_config(file: str, target: dict):
	y_dep = su.load_yaml(file)
	return {**target, **y_dep}


def _assure_dir(name: str, target_path: str, create: bool):
	if not os.path.isdir(target_path):
		if create:
			print(">", name, ":", target_path, "not found. Creating...")
			os.makedirs(target_path)
		else:
			print("! Not allowed to create" + name + ". Exit")
			sys.exit(1)


def assure_dirs():
	sg.configuration[C_TEX_HOME] = su.get_sltx_tex_home()  # expansion
	create = sg.configuration[C_CREATE_DIRS]
	_assure_dir('Tex-Home', sg.configuration[C_TEX_HOME], create)

	for config, name in [(C_WORKING_DIR, 'Working-Dir'), (C_DOWNLOAD_DIR, 'Download-Dir'),
			(C_CACHE_DIR, 'Cache-Dir')]:
		sg.configuration[config] = os.path.expanduser(sg.configuration[config])  # expansion
		_assure_dir(name, sg.configuration[config], create)




import os  # list directory
import re
import shutil  # clean working dir
import sys
from concurrent import futures
from pathlib import Path

from importlib_resources import files

import sltxpkg.lithie.compile.cooker as cooker
from sltxpkg import dep, generate, util as su, config as sc, globals as sg
from sltxpkg.config import (assure_dirs, load_configuration,
							load_dependencies_config, write_to_log)
from sltxpkg.dep import install_dependencies
from sltxpkg.globals import (C_CLEANUP, C_CREATE_DIRS, C_DOWNLOAD_DIR,
							 C_DRIVER_LOG, C_DRIVER_PATTERNS, C_DRIVERS,
							 C_RECURSIVE, C_TEX_HOME, C_USE_DOCKER,
							 C_WORKING_DIR, DEFAULT_CONFIG)
from sltxpkg.lithie import commands as lithiecmd

import sltxpkg.data


def cmd_auto_setup():
	sc.assure_dirs()
	cleanse_caches()

	if sg.args.auto_deps:
		dep_path = str(files(sltxpkg.data).joinpath('sltx-dep.yaml'))
		sg.dependencies = load_dependencies_config(
			dep_path, sg.dependencies)
		install_dependencies()

	# TODO: allow option to install local dependencies too using the shipped sltx-dep?
	lithiecmd._install(sg.configuration[sg.C_DOCKER_PROFILE])


def cmd_dependency():
	if sg.args.deps is None or len(sg.args.deps) == 0:
		print("You must supply a dependency 'file'.")
		exit(1)

	for dep in sg.args.deps:
		# will extend the dict with 'new' ones
		# should work even better if sltx-source.yaml files are present in the targets
		sg.dependencies = load_dependencies_config(dep, sg.dependencies)

	assure_dirs()

	target = su.get_sltx_tex_home() if sg.args.local_path is None else sg.args.local_path
	install_dependencies(target=target)


def cmd_version():
	print("This is sltx, a simple latex helper-utility")
	print("Tex-Home:", su.get_tex_home())
	print("Default config location:", su.get_default_conf(),"(present:",str(os.path.isfile(su.get_default_conf()))+")")
	print("Local config location:", su.get_local_conf(),"(present:",str(os.path.isfile(su.get_local_conf()))+")")
	print("Version:", su.get_version())


def cmd_docker():
	lithiecmd.install()


def cmd_raw_compile():
	# install possible deps
	for dep in sg.args.extra_dependencies:
		# will extend the dict with 'new' ones
		# should work even better if sltx-source.yaml files are present in the targets
		sg.dependencies = load_dependencies_config(dep, sg.dependencies)
	# i know just writing withut len is more pythonic but i like it more if it is explicit
	if len(sg.args.extra_dependencies) > 0:
		texmf_home = su.get_tex_home()
		print("Insalling additional dependencies.")
		assure_dirs()
		install_dependencies(target=texmf_home)

	cooker.cook()
	# if no exception we went here compiling fine
	if sg.configuration[sg.C_CLEANUP]:
		sg.args.exclude_patterns = []
		sg.args.cleanse_all = False
		sg.args.cleanse_cache = False
		cmd_cleanse()

def cmd_compile():
	if(sg.configuration[C_USE_DOCKER]):
		print("Using docker to compile")
		lithiecmd.compile()
	else:
		print("Docker was disabled, using local compilation.")
		cmd_raw_compile()


def cmd_gen_gha():
	generate.generate()


def should_be_excluded(file: str):
	if sg.args.exclude_patterns is None:
		return False

	for exclude_pattern in sg.args.exclude_patterns:
		if re.match(exclude_pattern, file):
			return True
	return False


def cleanse_caches():
	# TODO: clean up .latexmkrc entries; not the whole file
	cache_dir = sg.configuration[sg.C_CACHE_DIR]
	if os.path.isdir(cache_dir):
		print("Cleaning all the caches... (" + cache_dir + ")")
		# avoids deleting the cache dir itself
		for root, dirs, files in os.walk(cache_dir):
			for name in files:
				os.remove(os.path.join(root, name))
			for name in dirs:
				shutil.rmtree(os.path.join(root, name))
	else:
		print("No caches \"" + cache_dir + "\" were found. Skipping...")


def cmd_cleanse():
	sc.assure_dirs()
	# Delete all current log files
	# TODO: make this dry. avoid specifying the log files signature multiple times (see Recipe)
	print("Cleaning local logs...")
	clean_patterns = ['sltx-log-*.tar.gz', 'sltx-log-*.zip', 'sltx-drivers.log']
	for clean_pattern in clean_patterns:
		for f in Path(".").glob(clean_pattern):
			if should_be_excluded(str(f)):
				print("File", f, "excluded.")
			else:
				f.unlink()
	if sg.args.cleanse_all:
		thome = su.get_sltx_tex_home()
		if os.path.isdir(thome):
			print("Cleaning sltx-texmf-tree... (" + thome + ")")
			shutil.rmtree(thome)
		else:
			print("The local sltx-texmf tree in \"" +
				  thome + "\" was not found. Skipping...")

	if sg.args.cleanse_all or sg.args.cleanse_cache:
		cleanse_caches()



# cookes the recipe :D
from concurrent import futures
import sys

import sltxpkg.globals as sg
from sltxpkg.lithie.compile.recipe import Recipe
import sltxpkg.lithie.compile.recipe_exceptions as rex

def cook():
	recipe_path = sg.configuration[sg.C_DEFAULT_RECIPE] if sg.args.recipe is None else sg.args.recipe

	with futures.ThreadPoolExecutor(max_workers=sg.args.threads) as pool:
		runners = []
		for i, file in enumerate(sg.args.files):
			recipe = Recipe(recipe_path, file, i)
			runners.append((pool.submit(recipe.run), i, file))
		futures.wait([r[0] for r in runners])
		try:
			for runner, i, file in runners:
				if runner.result() is not None:
					print("Status for runner:", i,"operating on file:",file)
					print(runner.result())
		except rex.RecipeException as ex:
			print("\n\033[31m ! Processing of",file,"failed for:",repr(ex),"\033[m")
			sys.exit(128)
		else:
			print("\n=Compiled all documents successfully=")



import sltxpkg.prompt as prompt
from sltxpkg.lithie.docker_mg import DockerCtrl
import sltxpkg.globals as sg
import sltxpkg.config as sc

# TODO: get valid names
def install():
	valid_profiles=['tx-small', 'tx-default', 'tx-full']
	print("Please enter the profile you want. Valid names are:", valid_profiles)
	target = prompt.get("Profile [{default}]", default=sg.configuration[sg.C_DOCKER_PROFILE]).lower()
	# TODO: do not exit, try anyway
	if target not in valid_profiles:
		print(target, "was not found in", valid_profiles, "exiting for now")
		exit(1)

	_install(target)

def _install(target: str):
	docker_ctrl = DockerCtrl()
	docker_ctrl.update_img(target)

# TODO: clarify time in docker container as it is utc it must be the same as host!!!!
# TODO: keep container running to avoid full recompilations each time? (this should be the optimum!!)
# TODO: optimal would be to mount the working dir to the real world outside working dir to have it detached but saved
# TODO: in this case cleaning would not be needed!
def compile():
	sc.assure_dirs()
	docker_ctrl = DockerCtrl()
	profile = sg.configuration[sg.C_DOCKER_PROFILE] if sg.args.profile is None else sg.args.profile
	sltx_command = "sltx -t " + str(sg.args.threads) + " "

	if sg.args.quiet:
		sltx_command += "--quiet "

	if sg.args.config is not None:
		sltx_command += "--config \"" + sg.args.config + "\" "

	sltx_command += "raw-compile "

	if sg.args.recipe is not None:
		sltx_command += "--recipe \"" + sg.args.recipe + "\" "

	if sg.args.extra_arguments is not None:
		sltx_command += "--args=\"" + " ".join(sg.args.extra_arguments) + "\" "

	for dep in sg.args.extra_dependencies:
		# will extend the dict with 'new' ones
		# should work even better if sltx-source.yaml files are present in the targets
		sltx_command += "--dependency \"" + dep + "\" "

	sltx_command += " ".join(['"' + f + '"' for f in sg.args.files])

	print("Running command in docker: " + sltx_command)
	docker_ctrl.run_in_container(sg.args.dock_as_root, profile, sltx_command)


import mmap
import os
from sys import platform
from time import localtime, strftime
import re

import yaml
from importlib_resources import as_file, files

import sltxpkg.data
import sltxpkg.globals as sg


def default_texmf():
	if platform == "linux" or platform == "linux2":
		return "~/texmf"
	elif platform == "darwin":
		return "~/Library/texmf"
	elif platform == "win32":
		return "~/texmf"


def get_version():
	return files(sltxpkg.data).joinpath('version.info').read_text()


def load_yaml(file_path: str):
	with open(file_path, 'r') as yaml_file:
		# FullLoader only available for 5.1 and above:
		if float(yaml.__version__[:yaml.__version__.rfind(".")]) >= 5.1:
			return yaml.load(yaml_file, Loader=yaml.FullLoader)
		else:
			return yaml.load(yaml_file)


def file_contains(path: str, txt: str):
	with open(path, 'rb', 0) as file, \
			mmap.mmap(file.fileno(), 0, access=mmap.ACCESS_READ) as s:
		return s.find(txt.encode('utf-8')) != -1


def get_now():
	return strftime("%Y-%m-%d__%H-%M-%S", localtime())


def get_default_conf() -> str:
	return os.path.expanduser(sg.DEFAULT_CONFIG)


def get_local_conf() -> str:
	return os.path.expanduser(sg.LOCAL_CONFIG)


def get_tex_home() -> str:
	return os.path.expanduser(default_texmf())


def get_sltx_tex_home() -> str:
	return os.path.expanduser(sg.configuration[sg.C_TEX_HOME].format(
		**sg.configuration, os_default_texmf=default_texmf()))

# For compile


def sanitize_filename(text: str):
	return re.sub('[^a-zA-Z0-9\-_]', '_', text)



import sltxpkg.prompt as prompt
import os

import yaml
from yaml.representer import SafeRepresenter

import sltxpkg.globals as sg
from sltxpkg.globals import C_TEX_HOME

# https://stackoverflow.com/a/20863889


def change_style(style, representer):
	def new_representer(dumper, data):
		scalar = representer(dumper, data)
		scalar.style = style
		return scalar
	return new_representer


class YamlBlock(str):
	pass


represent_literal_str = change_style('|', SafeRepresenter.represent_str)
yaml.add_representer(YamlBlock, represent_literal_str)


def assure_workflow_target(path: str):
	if os.path.isfile(path):
		print("A workflow-file with the given name does already exist!")
		overwrite = prompt.get_bool(default=False)
		if not overwrite:
			print("Aborting...")
			exit(1)
	basepath = os.path.dirname(path)
	if basepath is None or basepath.strip() == "":
		return

	if not os.path.isdir(basepath):
		print("The directory", basepath, "does not exist. should it be created?")
		create = prompt.get_bool(default=True)
		if not create:
			print("Aborting...")
			exit(1)
		os.makedirs(basepath)


def add_step(document: dict, name: str, uses: str = None, _with: dict = None, _run: str = None):
	step = {'name': name}
	if uses is not None:
		step['uses'] = uses
	if _with is not None:
		step['with'] = _with
	if _run is not None:
		step['run'] = _run
	document['jobs']['build']['steps'].append(step)


def step_checkout(document: dict):
	lfs = prompt.get_bool("Checkout lfs ", default=True)
	submodules = prompt.get_bool("Checkout submodules ", default=True)
	add_step(document,
			 "Checkout git repository",
			 uses="actions/checkout@v2",
			 _with={
				 'lfs': lfs,
				 'submodules': submodules
			 })


def step_setup_python(document: dict):
	add_step(document,
			 "Setup Python environment for sltx",
			 uses="actions/setup-python@v2",
			 _with={
				 'python-version': '3.8'
			 })


def step_setup_sltx(document: dict) -> str:
	# tODO get them with 'docker search' # TODO dry
	valid_profiles = ['tx-small', 'tx-default', 'tx-full']
	print("Please enter the profile you want for sltx. Valid names are:", valid_profiles)
	target_profile = prompt.get(
		"Profile [{default}]", default=sg.configuration[sg.C_DOCKER_PROFILE]).lower()

	# TODO: we need a option for additional dependencies
	setup_lines = "echo \"" + target_profile + "\" | sltx docker"

	add_step(document,
			 "Setup and run sltx-install",
			 _run=YamlBlock("pip install sltx\n" +
							setup_lines + "\n")
			 )
	return target_profile


def step_compile(document: dict, target_profile: str):
	print("Do you need your own config file to compile?")
	own_config = prompt.get_bool(default=False)
	config_file = None
	if own_config:
		config_file = prompt.get_file("Path to config-file")

	print("Which documents do you want to have compiled? You may separate multiple ones by comma.")
	file_list = prompt.get("File(s)")
	files = file_list.split(',')

	exec_line = "sltx "

	# TODO: fail as soon as one fails
	if own_config and config_file is not None:
		exec_line += "--config \"{config_file}\" ".format(**locals())

	# TODO: recipe support!
	# NOTE: root is required for now due to permission errors in action container
	# this is only for gha
	exec_line += "compile --root --profile \"" + target_profile + "\" "  + " ".join(['"' + f + '"' for f in files]) + "\n"

	add_step(document,
			 "Compile the Documents",
			 _run=YamlBlock(exec_line)
			 )
	return files


def step_commit_and_push(document: dict, files: list):
	print("Do you want the resulting pdf(s) to be pushed?")
	do_push = prompt.get_bool(default=True)

	if not do_push:
		return

	files = [os.path.splitext(f)[0] + ".pdf" for f in files]

	print("Which documents do you want to have pushed? You may separate multiple ones by comma.")
	print("Default:", files)
	push_file_list = prompt.get("Doc(s)", default="")
	if push_file_list.strip() != "":
		files = push_file_list.split(',')

	add_line = "git add -f " + \
		" ".join(["\"" + f + "\"" for f in files]) + "\n"

	# commit
	add_step(document,
			 "Commit",
			 _run=YamlBlock(
				 "git config --local user.email \"action@github.com\"\ngit config --local user.name \"GitHub Action\"\n" +
				 add_line + "git commit -m \"Newly compiled data\"\n"))

	# push
	branch = prompt.get("Push-Branch [{default}]", default="gh-pages")
	add_step(document,
			 "Push changes",
			 uses="ad-m/github-push-action@master",
			 _with={
				 'branch': branch,
				 'github_token': "${{ secrets.GITHUB_TOKEN }}",
				 'force': True
			 })


def generate():
	document = {}

	print("We will now generate a GitHub-Action workflow")
	target_path = prompt.get(
		"Workflow-Path [{default}]", default=".github/workflows/compile.yaml")
	assure_workflow_target(target_path)

	document['name'] = prompt.get("Workflow name")
	document['on'] = {'push': {'branches': ['master','main']}}
	document['jobs'] = {
		'build': {
			'runs-on': 'ubuntu-latest',
			'steps': []
		}
	}

	step_checkout(document)
	step_setup_python(document)
	profile = step_setup_sltx(document)
	files = step_compile(document, profile)
	step_commit_and_push(document, files)

	print("Ok, I will write the file now...")
	with open(target_path, 'tw') as f:
		# We will disable this to get name on top and have no sorting
		stream = yaml.dump(document, default_flow_style=False, sort_keys=False)
		# formatting a little bit?
		f.write(stream.replace('jobs:', '\njobs:'))
	print("File written to \"" + target_path + "\". Job completed.")



import setuptools

with open('README.md', 'r') as readme:
	long_description = readme.read()

with open('sltxpkg/data/version.info', 'r') as fv:
	version = fv.readline()

setuptools.setup(
	name="sltx",
	version=version,
	author="Florian Sihler",
	author_email="florian.sihler@uni-ulm.de",
	description="sltx-utility",
	long_description=long_description,
	long_description_content_type="text/markdown",
	install_requires=['PyYAML', 'docker', 'importlib_resources'],
	scripts=['sltx'],
	url="https://github.com/EagleoutIce/sltx",
	packages=setuptools.find_packages(),
	package_data={
		'data': ['sltxpkg/data/version.info', 'sltxpkg/data/sltx-dep.yaml'],
		'data/recipes': ['sltxpkg/data/recipes/default-latexmk.recipe','sltxpkg/data/recipes/compress-latexmk.recipe'],
		'data/latexmk/configs': ['data/latexmk/configs/glossary.mkrc']
	},
	include_package_data=True,
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
		'Intended Audience :: Developers',
		'Environment :: Console'
	],
	python_requires='>=3.5',
)


import os

def valid_default(inp: str):
	return inp is not None and inp.strip() != ""


def valid_dir(inp: str):
	return valid_default(inp) and os.path.isdir(inp)


def valid_file(inp: str):
	return valid_default(inp) and os.path.isfile(inp)


def get(prompt: str, valid_input=valid_default, type=str, default=None):
	got = None
	prompt = "\033[36m" + prompt.format(**locals()) + ">\033[m "
	if default is None:
		while not valid_input(got):
			got = input(prompt)
	else:
		got = input(prompt)
	ret = type(default if not valid_input(got) else got)
	return ret


def get_bool(pre="", default=None):
	if default is None:
		prompt = "[true/false]"
	elif default:
		prompt = "[TRUE/false]"
	else:
		prompt = "[True/FALSE]"

	return get(pre + prompt, type=bool, default=default)


def get_dir(prompt: str, default=None):
	return get(prompt, valid_input=valid_dir, type=str, default=default)


def get_file(prompt: str, default=None):
	return get(prompt, valid_input=valid_file, type=str, default=default)